import threading
from typing import Dict, Tuple

from cv2 import VideoWriter

from je_auto_control.wrapper.auto_control_screen import screenshot


class ScreenRecorder(object):

    def __init__(self):
        self.running_recorder: Dict[str, ScreenRecordThread] = {}

    def start_new_recode(self, recoder_name: str, path_and_filename: str = "output.avi", codec: str = "XVID",
                         frame_per_sec: int = 30, resolution: Tuple[int, int] = (1920, 1080)):
        record_thread = ScreenRecordThread(path_and_filename, codec, frame_per_sec, resolution)
        old_record = self.running_recorder.get(recoder_name, None)
        if old_record is not None:
            old_record.record_flag = False
        record_thread.daemon = True
        record_thread.start()
        self.running_recorder.update({recoder_name: record_thread})


class ScreenRecordThread(threading.Thread):

    def __init__(self, path_and_filename, codec, frame_per_sec, resolution: Tuple[int, int]):
        super().__init__()
        self.fourcc = VideoWriter.fourcc(*codec)
        self.video_writer = VideoWriter(path_and_filename, self.fourcc, frame_per_sec, resolution)
        self.record_flag = False

    def run(self) -> None:
        self.record_flag = True
        while self.record_flag:
            # Get raw pixels from the screen, save it to a Numpy array
            image = screenshot()
            self.video_writer.write(image)
        else:
            self.video_writer.release()



from typing import List

from PIL import ImageGrab
from je_open_cv import template_detection


def find_image(image, detect_threshold: float = 1, draw_image: bool = False) -> List[int]:
    """
    Find image with detect threshold on screen.
    :param image: which cv2_utils we want to find on screen
    :param detect_threshold: detect precision 0.0 ~ 1.0; 1 is absolute equal
    :param draw_image: draw detect tag on return cv2_utils
    """
    grab_image = ImageGrab.grab()
    return template_detection.find_object(image=grab_image, template=image,
                                          detect_threshold=detect_threshold, draw_image=draw_image)


def find_image_multi(image, detect_threshold: float = 1, draw_image: bool = False) -> List[List[int]]:
    """
    Find multi image with detect threshold on screen.
    :param image: which cv2_utils we want to find on screen
    :param detect_threshold: detect precision 0.0 ~ 1.0; 1 is absolute equal
    :param draw_image: draw detect tag on return cv2_utils
    """
    grab_image = ImageGrab.grab()
    return template_detection.find_multi_object(image=grab_image, template=image,
                                                detect_threshold=detect_threshold, draw_image=draw_image)



from multiprocessing import Process

from je_auto_control.utils.exception.exception_tags import timeout_need_on_main_error
from je_auto_control.utils.exception.exceptions import AutoControlTimeoutException


def multiprocess_timeout(check_function, time: int) -> str:
    try:
        new_process: Process = Process(target=check_function)
        new_process.start()
        new_process.join(timeout=time)
    except AutoControlTimeoutException:
        raise AutoControlTimeoutException(timeout_need_on_main_error)
    new_process.terminate()
    if new_process.exitcode is None:
        return "timeout"
    else:
        return "success"



import sys

from je_auto_control.utils.exception.exception_tags import macos_record_error
from je_auto_control.utils.exception.exceptions import AutoControlException
from je_auto_control.utils.exception.exceptions import AutoControlJsonActionException
from je_auto_control.utils.logging.loggin_instance import autocontrol_logger
from je_auto_control.utils.test_record.record_test_class import record_action_to_list
from je_auto_control.wrapper.platform_wrapper import recorder


def record() -> None:
    """
    start record keyboard and mouse event until stop_record
    """
    autocontrol_logger.info("record")
    try:
        if sys.platform == "darwin":
            raise AutoControlException(macos_record_error)
        record_action_to_list("record", None)
        recorder.record()
    except Exception as error:
        record_action_to_list("record", None, repr(error))
        autocontrol_logger.error(f"record, failed: {repr(error)}")


def stop_record() -> list:
    """
    stop current record
    """
    autocontrol_logger.info("stop_record")
    try:
        if sys.platform == "darwin":
            raise AutoControlException(macos_record_error)
        action_queue = recorder.stop_record()
        if action_queue is None:
            raise AutoControlJsonActionException
        action_list = list(action_queue.queue)
        new_list = list()
        for action in action_list:
            if action[0] == "AC_type_keyboard":
                new_list.append([action[0], dict([["keycode", action[1]]])])
            else:
                new_list.append([action[0], dict(zip(["mouse_keycode", "x", "y"], [action[0], action[1], action[2]]))])
        record_action_to_list("stop_record", None)
        return new_list
    except Exception as error:
        record_action_to_list("stop_record", None, repr(error))
        autocontrol_logger.error(f"stop_record, failed: {repr(error)}")



import sys

from je_auto_control.utils.exception.exception_tags import windows_import_error
from je_auto_control.utils.exception.exceptions import AutoControlException

if sys.platform not in ["win32", "cygwin", "msys"]:
    raise AutoControlException(windows_import_error)

from je_auto_control.windows.listener.win32_keyboard_listener import Win32KeyboardListener
from je_auto_control.windows.listener.win32_mouse_listener import Win32MouseListener

from queue import Queue


class Win32Recorder(object):

    def __init__(self):
        self.mouse_record_listener: [None, Win32MouseListener] = None
        self.keyboard_record_listener: [None, Win32KeyboardListener] = None
        self.record_queue: [None, Queue] = None
        self.result_queue: [None, Queue] = None

    def record(self) -> None:
        self.mouse_record_listener = Win32MouseListener()
        self.keyboard_record_listener = Win32KeyboardListener()
        self.record_queue = Queue()
        self.mouse_record_listener.record(self.record_queue)
        self.keyboard_record_listener.record(self.record_queue)

    def stop_record(self) -> Queue:
        self.result_queue = self.mouse_record_listener.stop_record()
        self.result_queue = self.keyboard_record_listener.stop_record()
        self.record_queue = None
        return self.result_queue

    def record_mouse(self) -> None:
        self.mouse_record_listener = Win32MouseListener()
        self.record_queue = Queue()
        self.mouse_record_listener.record(self.record_queue)

    def stop_record_mouse(self) -> Queue:
        self.result_queue = self.mouse_record_listener.stop_record()
        self.record_queue = None
        return self.result_queue

    def record_keyboard(self) -> None:
        self.keyboard_record_listener = Win32KeyboardListener()
        self.record_queue = Queue()
        self.keyboard_record_listener.record(self.record_queue)

    def stop_record_keyboard(self) -> Queue:
        self.result_queue = self.keyboard_record_listener.stop_record()
        self.record_queue = None
        return self.result_queue


win32_recorder = Win32Recorder()



import sys
from typing import Tuple

from je_auto_control.utils.exception.exception_tags import keyboard_hotkey
from je_auto_control.utils.exception.exception_tags import keyboard_press_key
from je_auto_control.utils.exception.exception_tags import keyboard_release_key
from je_auto_control.utils.exception.exception_tags import keyboard_type_key
from je_auto_control.utils.exception.exception_tags import keyboard_write
from je_auto_control.utils.exception.exception_tags import keyboard_write_cant_find
from je_auto_control.utils.exception.exception_tags import table_cant_find_key
from je_auto_control.utils.exception.exceptions import AutoControlCantFindKeyException
from je_auto_control.utils.exception.exceptions import AutoControlKeyboardException
from je_auto_control.utils.logging.loggin_instance import autocontrol_logger
from je_auto_control.utils.test_record.record_test_class import record_action_to_list
from je_auto_control.wrapper.platform_wrapper import keyboard, special_mouse_keys_table
from je_auto_control.wrapper.platform_wrapper import keyboard_check
from je_auto_control.wrapper.platform_wrapper import keyboard_keys_table


def get_special_table() -> dict:
    return special_mouse_keys_table


def get_keyboard_keys_table() -> dict:
    return keyboard_keys_table


def press_keyboard_key(keycode: [int, str], is_shift: bool = False, skip_record: bool = False) -> str:
    """
    use to press a key still press to use release key
    or use critical exit
    return keycode
    :param keycode which keycode we want to press
    :param is_shift press shift True or False
    :param skip_record skip record on record total list True or False
    """
    autocontrol_logger.info(
        f"press_keyboard_key, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}"
    )
    param = locals()
    try:
        if isinstance(keycode, str):
            try:
                keycode = keyboard_keys_table.get(keycode)
            except AutoControlCantFindKeyException:
                autocontrol_logger.error(
                    f"press_keyboard_key failed, keycode: {keycode}, is_shift: {is_shift}, "
                    f"failed: {repr(AutoControlCantFindKeyException(table_cant_find_key))}"
                )
                raise AutoControlCantFindKeyException(table_cant_find_key)
        try:
            if sys.platform in ["win32", "cygwin", "msys", "linux", "linux2"]:
                keyboard.press_key(keycode)
            elif sys.platform in ["darwin"]:
                keyboard.press_key(keycode, is_shift=is_shift)
            if skip_record is False:
                record_action_to_list("press_key", param)
            return str(keycode)
        except AutoControlKeyboardException as error:
            if skip_record is False:
                record_action_to_list("press_key", param, repr(error))
            autocontrol_logger.error(
                f"press_keyboard_key failed, keycode: {keycode}, is_shift: {is_shift}, "
                f"{repr(AutoControlKeyboardException(keyboard_press_key + ' ' + repr(error)))}"
            )
            raise AutoControlKeyboardException(keyboard_press_key + " " + repr(error))
        except TypeError as error:
            if skip_record is False:
                record_action_to_list("press_key", param, repr(error))
            autocontrol_logger.error(
                f"press_keyboard_key failed, keycode: {keycode}, is_shift: {is_shift}, "
                f"failed: {repr(AutoControlKeyboardException)}"
            )
            raise AutoControlKeyboardException(repr(error))
    except Exception as error:
        if skip_record is False:
            record_action_to_list("press_key", param, repr(error))
        autocontrol_logger.error(
            f"press_keyboard_key failed, keycode: {keycode}, is_shift: {is_shift}, "
            f"failed: {repr(error)}"
        )


def release_keyboard_key(keycode: [int, str], is_shift: bool = False, skip_record: bool = False) -> str:
    """
    use to release pressed key return keycode
    :param keycode which keycode we want to release
    :param is_shift press shift True or False
    :param skip_record skip record on record total list True or False
    """
    autocontrol_logger.info(
        f"release_keyboard_key, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}"
    )
    param = locals()
    try:
        if isinstance(keycode, str):
            try:
                keycode = keyboard_keys_table.get(keycode)
            except AutoControlCantFindKeyException:
                raise AutoControlCantFindKeyException(table_cant_find_key)
        try:
            if sys.platform in ["win32", "cygwin", "msys", "linux", "linux2"]:
                keyboard.release_key(keycode)
            elif sys.platform in ["darwin"]:
                keyboard.release_key(keycode, is_shift=is_shift)
            if skip_record is False:
                record_action_to_list("release_key", param)
            return str(keycode)
        except AutoControlKeyboardException as error:
            if skip_record is False:
                record_action_to_list("release_key", param, repr(error))
            autocontrol_logger.error(
                f"release_keyboard_key, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}, "
                f"failed: {AutoControlKeyboardException(keyboard_release_key + ' ' + repr(error))}"
            )
            raise AutoControlKeyboardException(keyboard_release_key + " " + repr(error))
        except TypeError as error:
            if skip_record is False:
                record_action_to_list("release_key", param, repr(error))
            autocontrol_logger.error(
                f"release_keyboard_key, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}, "
                f"failed: {AutoControlKeyboardException(error)}"
            )
            raise AutoControlKeyboardException(error)
    except Exception as error:
        if skip_record is False:
            record_action_to_list("release_key", param, repr(error))
        autocontrol_logger.error(
            f"release_keyboard_key, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}, "
            f"failed: {repr(error)}"
        )


def type_keyboard(keycode: [int, str], is_shift: bool = False, skip_record: bool = False) -> str:
    """
    press and release key return keycode
    :param keycode which keycode we want to type
    :param is_shift press shift True or False
    :param skip_record skip record on record total list True or False
    """
    autocontrol_logger.info(
        f"type_keyboard, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}"
    )
    param = locals()
    try:
        try:
            press_keyboard_key(keycode, is_shift, skip_record=True)
            release_keyboard_key(keycode, is_shift, skip_record=True)
            if skip_record is False:
                record_action_to_list("type_keyboard", param)
            return str(keycode)
        except AutoControlKeyboardException as error:
            if skip_record is False:
                record_action_to_list("type_keyboard", param, repr(error))
            autocontrol_logger.error(
                f"type_keyboard, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}, "
                f"failed: {repr(AutoControlKeyboardException(keyboard_type_key + ' ' + repr(error)))}"
            )
            raise AutoControlKeyboardException(keyboard_type_key + " " + repr(error))
        except TypeError as error:
            if skip_record is False:
                record_action_to_list("type_keyboard", param, repr(error))
            autocontrol_logger.error(
                f"type_keyboard, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}, "
                f"failed: {repr(AutoControlKeyboardException(repr(error)))}"
            )
            raise AutoControlKeyboardException(repr(error))
    except Exception as error:
        if skip_record is False:
            record_action_to_list("type_keyboard", param, repr(error))
        autocontrol_logger.error(
            f"type_keyboard, keycode: {keycode}, is_shift: {is_shift}, skip_record: {skip_record}, "
            f"failed: {repr(error)}"
        )


def check_key_is_press(keycode: [int, str]) -> bool:
    """
    use to check key is press return True or False
    :param keycode check key is press or not
    """
    autocontrol_logger.info(
        f"check_key_is_press, keycode: {keycode}"
    )
    param = locals()
    try:
        if isinstance(keycode, int):
            get_key_code = keycode
        else:
            get_key_code = keyboard_keys_table.get(keycode)
        record_action_to_list("check_key_is_press", param)
        return keyboard_check.check_key_is_press(keycode=get_key_code)
    except Exception as error:
        record_action_to_list("check_key_is_press", param, repr(error))
        autocontrol_logger.error(
            f"check_key_is_press, keycode: {keycode}, "
            f"failed: {repr(error)}"
        )


def write(write_string: str, is_shift: bool = False) -> str:
    """
    use to press and release whole we get this function str
    return all press and release str
    :param write_string while string not on write_string+1 type_keyboard(string)
    :param is_shift press shift True or False
    """
    autocontrol_logger.info(
        f"write, write_string: {write_string}, is_shift: {is_shift}"
    )
    param = locals()
    try:
        try:
            record_write_string = ""
            for single_string in write_string:
                try:
                    key = keyboard_keys_table.get(single_string, None)
                    if key is not None:
                        record_write_string = "".join(
                            [
                                record_write_string,
                                type_keyboard(key, is_shift, skip_record=True)
                            ]
                        )
                    elif single_string.isspace():
                        record_write_string = "".join(
                            [
                                record_write_string,
                                type_keyboard("space", is_shift, skip_record=True)
                            ]
                        )
                    else:
                        autocontrol_logger.error(
                            f"write, write_string: {write_string}, is_shift: {is_shift}, "
                            f"failed: {AutoControlKeyboardException(keyboard_write_cant_find)}"
                        )
                        raise AutoControlKeyboardException(keyboard_write_cant_find)
                except AutoControlKeyboardException as error:
                    autocontrol_logger.error(
                        f"write, write_string: {write_string}, is_shift: {is_shift}, "
                        f"failed: {repr(error)}, keyboard_write_cant_find, {single_string}"
                    )
                    raise AutoControlKeyboardException(keyboard_write_cant_find)
            record_action_to_list("write", param)
            return record_write_string
        except AutoControlKeyboardException as error:
            autocontrol_logger.error(
                f"write, write_string: {write_string}, is_shift: {is_shift}, "
                f"failed: {AutoControlKeyboardException(keyboard_write + ' ' + repr(error))}"
            )
            raise AutoControlKeyboardException(keyboard_write + " " + repr(error))
    except Exception as error:
        record_action_to_list("write", param, repr(error))
        autocontrol_logger.error(
            f"write, write_string: {write_string}, is_shift: {is_shift}, "
            f"failed: {repr(error)}"
        )


def hotkey(key_code_list: list, is_shift: bool = False) -> Tuple[str, str]:
    """
    use to press and release all key on key_code_list
    then reverse list press and release again
    return [press_str_list, release_str_list]
    :param key_code_list press and release all key on list and reverse
    :param is_shift press shift True or False
    """
    autocontrol_logger.info(
        f"hotkey, key_code_list: {key_code_list}, is_shift: {is_shift}"
    )
    param = locals()
    try:
        try:
            record_hotkey_press_string = ""
            record_hotkey_release_string = ""
            for key in key_code_list:
                record_hotkey_press_string = ",".join(
                    [
                        record_hotkey_press_string,
                        press_keyboard_key(key, is_shift, skip_record=True)
                    ]
                )
            key_code_list.reverse()
            for key in key_code_list:
                record_hotkey_release_string = ",".join(
                    [
                        record_hotkey_release_string,
                        release_keyboard_key(key, is_shift, skip_record=True)
                    ]
                )
            record_action_to_list("hotkey", param)
            return record_hotkey_press_string, record_hotkey_release_string
        except AutoControlKeyboardException as error:
            autocontrol_logger.error(
                f"hotkey, key_code_list: {key_code_list}, is_shift: {is_shift}, "
                f"failed: {AutoControlKeyboardException(keyboard_hotkey + ' ' + repr(error))}"
            )
            raise AutoControlKeyboardException(keyboard_hotkey + " " + repr(error))
    except Exception as error:
        record_action_to_list("hotkey", param, repr(error))
        autocontrol_logger.error(
            f"hotkey, key_code_list: {key_code_list}, is_shift: {is_shift}, "
            f"failed: {repr(error)}"
        )
